import base64
import json


class BaseSessionBackend:
    def encode(self, session: dict) -> str:
        return base64.b64encode(json.dumps(session)).decode()

    def decode(self, data: str) -> dict:
        return json.loads(base64.b64decode(data.encode()))




from fastapi import FastAPI
from joey import db, utils


class Application(FastAPI):
    def __init__(self, settings, *args, **kwargs):
        super().__init__(*args, **kwargs)
        db.init(settings)
        for application_name, properties in settings.ROUTES.items():
            self.include_router(utils.get_router(application_name), **properties)
        for middleware_name, properties in settings.MIDDLEWARE.items():
            self.add_middleware(utils.get_middleware(middleware_name), **properties)



from typing import List, Dict

from pydantic import BaseSettings


def variables(module):
    return {k: v for k, v in vars(dev).items() if not k.startswith('_') and k.isupper()}


class Settings(BaseSettings):
    PROJECT_NAME: str
    SQLALCHEMY_DATABASE_URI: str
    APPLICATIONS: List[str]
    ROUTES: Dict[str, dict]
    MIDDLEWARE: Dict[str, dict]


try:
    import settings.local as local

    settings = Settings(**variables(local))
except ImportError:
    import settings.development as dev

    settings = Settings(**variables(dev))



import logging
import subprocess
import sys
import uuid
from pathlib import Path
from functools import wraps

import click

from .utils import copy_directory_structure, register_app

logger = logging.getLogger('cli')
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)


def verbose(func):
    @wraps(func)
    def wrapper(verbose, *args, **kwargs):
        if verbose:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
        return func(*args, **kwargs)

    return click.option('--verbose', '-v', is_flag=True, default=False, help='Verbose output')(wrapper)


@click.group()
@verbose
def cli():
    '''Async web framework on top of fastapi and orm'''
    pass


@cli.command()
@click.argument('name', required=False)
@verbose
def init(name=None):
    '''Create project skeleton'''
    app_directory = Path(name) if name else Path('.')
    variables = {
        'app_name': name if name else app_directory.cwd().name,
        'secret_key': uuid.uuid4().hex,
        'cookie_secret': uuid.uuid4().hex,
    }
    templates_dir = Path(__file__).parent / 'templates' / 'project'

    logger.debug('Start project creation')
    copy_directory_structure(templates_dir, app_directory, variables)
    logger.info('Project `{app_name}` successfully created'.format(**variables))


@cli.command()
@click.argument('name')
@click.option('--autoregister', '-a', is_flag=True, default=False, help='Autoregister application')
@verbose
def add(name, autoregister):
    '''Add application to current project'''
    app_directory = Path(name)

    if app_directory.exists():
        logger.error(f'Application folder with name `{name}` already exist')
        return

    templates_dir = Path(__file__).parent / 'templates' / 'app'
    logger.debug('Start application creation')
    copy_directory_structure(templates_dir, app_directory, {})

    if autoregister:
        config_file = Path('.') / 'settings' / 'common.yml'
        if config_file.exists():
            register_app(name, config_file)

    logger.info(f'Application `{name}` successfully added')


@cli.command()
@verbose
def run():
    '''Run developer server'''
    try:
        subprocess.call(['uvicorn', '--reload', 'asgi:app'])
    except FileNotFoundError:
        logger.error('Uvicorn is not installed. Run asgi:app with your favorite async server.')
    except KeyboardInterrupt:
        pass


@cli.command()
@click.argument('label', required=False)
@verbose
def revise(label='auto'):
    '''Create migration'''
    subprocess.call(['alembic', 'revision', '-m', label, '--autogenerate'])


@cli.command()
@verbose
def migrate():
    '''Apply migration'''
    subprocess.call(['alembic', 'upgrade', 'head'])


def main(argv=None):
    cli(args=argv)



import sys
from importlib import import_module
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

config = context.config
fileConfig(config.config_file_name)

# add parent dir to python path for proper imports
sys.path += ['.']
from joey.db import Model, init
from settings import settings
init(settings)
target_metadata = Model.__metadata__

# import all models to fill metadata
for app in settings.APPLICATIONS:
    import_module(app + ".models")


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """

    url = settings.SQLALCHEMY_DATABASE_URI
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        {"sqlalchemy.url": settings.SQLALCHEMY_DATABASE_URI}, prefix="sqlalchemy.", poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata, render_as_batch=True)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import datetime

from .base import BaseSessionBackend

try:
    import jwt
    from cryptography.fernet import Fernet
except ImportError as e:
    print('PyJWT and cryptography are required. Install them or joey[sessions].')
    raise e


class FernetSessionBackend(BaseSessionBackend):
    def __init__(self, cipher_key, signature_key=None, max_age=datetime.timedelta(weeks=2).total_seconds()):
        self.fernet = Fernet(cipher_key)
        self.signature_key = signature_key if signature_key else cipher_key
        self.max_age = max_age

    def encode(self, session: dict) -> str:
        return self.fernet.encrypt(
            jwt.encode(
                {'exp': datetime.datetime.utcnow() + datetime.timedelta(seconds=self.max_age), **session},
                self.signature_key,
                algorithm='HS256',
            ).encode()
        ).hex

    def decode(self, data: str) -> dict:
        try:
            data = self.fernet.decrypt(bytes.fromhex(data))
            return jwt.decode(data, self.signature_key, algorithms=['HS256'])
        except jwt.ExpiredSignatureError:
            return {}



from pathlib import Path
import logging

import yaml
from mako.exceptions import RichTraceback
from mako.template import Template

logger = logging.getLogger('cli')

def _render(text: str, **kwargs) -> str:
    if kwargs:
        try:
            template = Template(text)
            return template.render(**kwargs)
        except:
            traceback = RichTraceback()
            errors = ['Template rendering error']
            for (filename, lineno, function, line) in traceback.traceback:
                errors.append(f'  file {filename}, line {lineno}, in {function}\n    {line}')
            errors.append(f'  {str(traceback.error.__class__.__name__)}: {traceback.error}')
            logger.error('\n'.join(errors))
            exit(-1)
    return text


def _is_renderable_template(path: Path) -> bool:
    return path.suffix == '.pyt'


def _redered_path(template_path: Path) -> Path:
    return template_path.with_suffix('.py')


def copy_directory_structure(src: Path, dst: Path, context: dict):
    for path in src.glob('**/*'):
        if path.is_dir() or path.suffix == '.pyc':
            continue

        text = path.open().read()
        if _is_renderable_template(path):
            text = _render(text, **context)
            path = _redered_path(path)

        filepath = dst / path.relative_to(src)
        filepath.parent.mkdir(parents=True, exist_ok=True)

        logger.debug(f' - {filepath}')
        with filepath.open('w') as f:
            f.write(text)


def register_app(name: str, config_file: Path):
    try:
        config = yaml.load(config_file.open(), Loader=yaml.Loader)
        applications = config['APPLICATIONS']
        routes = config['ROUTES']

        applications.append(name)
        routes[name] = {'prefix': f'/{name}', 'tags': [name]}

        with config_file.open('w') as f:
            yaml.dump(config, f)
        logger.info(f'Autoregister application and routes in `{config_file}`')
    except KeyError as e:
        logger.error('Invalid file format: key %s does not exist' % e.args)
    except (TypeError, AttributeError) as e:
        logger.error('Invalid file format')
    except Exception as e:
        logger.error(getattr(e, 'message', repr(e)))



import datetime
import typing

from starlette.datastructures import MutableHeaders, Secret
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from .backend.base import BaseSessionBackend

class SessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        backend: typing.Optional[BaseSessionBackend],
        session_cookie: str = 'session',
        max_age: int = datetime.timedelta(weeks=2).total_seconds(),  # 14 days, in seconds
        same_site: str = 'lax',
        https_only: bool = False,
    ) -> None:
        self.app = app
        self.session_cookie = session_cookie
        self.backend = backend if backend else BaseSessionBackend()
        self.max_age = max_age
        self.security_flags = 'httponly; samesite=' + same_site
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += '; secure'



    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['type'] not in ('http', 'websocket'):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        initial_session_was_empty = True

        session_data = connection.cookies.get(self.session_cookie, {})
        if session_data:
            session_data = self.backend.decode(session_data)
            if session_data:
                initial_session_was_empty = False
        scope['session'] = session_data

        async def send_wrapper(message: Message) -> None:
            if message['type'] == 'http.response.start':
                if scope['session']:
                    # We have session data to persist.
                    headers = MutableHeaders(scope=message)
                    header_value = '%s=%s; path=/; Max-Age=%d; %s' % (
                        self.session_cookie,
                        self.backend.encode(scope['session']),
                        self.max_age,
                        self.security_flags,
                    )
                    headers.append('Set-Cookie', header_value)
                elif not initial_session_was_empty:
                    # The session has been cleared.
                    headers = MutableHeaders(scope=message)
                    header_value = '%s=%s; %s' % (
                        self.session_cookie,
                        'null; path=/; expires=Thu, 01 Jan 1970 00:00:00 GMT;',
                        self.security_flags,
                    )
                    headers.append('Set-Cookie', header_value)
            await send(message)

        await self.app(scope, receive, send_wrapper)
